from dataclasses import dataclass, field
from typing import Union, Optional, Dict

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike
import numpy as np

from .float_types import FloatTypes
from .int_types import IntTypes
from .pixel_types import PixelTypes


@dataclass
class PolyDataType:
    pointPixelComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    pointPixelType: PixelTypes = PixelTypes.Scalar
    pointPixelComponents: int = 1

    cellPixelComponentType: Union[IntTypes, FloatTypes] = FloatTypes.Float32
    cellPixelType: PixelTypes = PixelTypes.Scalar
    cellPixelComponents: int = 1


def _default_points() -> ArrayLike:
    return np.empty((0,), np.float32)


@dataclass
class PolyData:
    polyDataType: Union[PolyDataType, Dict] = field(default_factory=PolyDataType)
    name: str = "PolyData"

    numberOfPoints: int = 0
    points: ArrayLike = field(default_factory=_default_points)

    verticesBufferSize: int = 0
    vertices: Optional[ArrayLike] = None

    linesBufferSize: int = 0
    lines: Optional[ArrayLike] = None

    polygonsBufferSize: int = 0
    polygons: Optional[ArrayLike] = None

    triangleStripsBufferSize: int = 0
    triangleStrips: Optional[ArrayLike] = None

    numberOfPointPixels: int = 0
    pointData: Optional[ArrayLike] = None

    numberOfCellPixels: int = 0
    cellData: Optional[ArrayLike] = None

    def __post_init__(self):
        if isinstance(self.polyDataType, dict):
            self.polyDataType = PolyDataType(**self.polyDataType)



from typing import Callable, Tuple, Dict, Optional, Set
import sys
import importlib
import sys

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


class FunctionFactory:
    def __init__(self):
        self._registered: Dict[Tuple[str, str], Set[Callable]] = {}
        self._priorities: Dict[Callable, int] = {}
        self._has_entry_point_lookup: Set[Tuple[str, str]] = set()

    def register(self, interface_package: str, func_name: str, func: Callable, priority: int = 1) -> None:
        key = (interface_package, func_name)
        registered = self._registered.get(key, set())
        registered.add(func)
        self._registered[key] = registered
        self._priorities[func] = priority

    def lookup(self, interface_package: str, func_name: str) -> Optional[Set[Callable]]:
        key = (interface_package, func_name)
        if not key in self._has_entry_point_lookup:
            discovered_funcs = entry_points(group=f"{interface_package}.{func_name}")
            for ep in discovered_funcs:
                priority = ep.name.partition(".priority.")[2]
                if priority:
                    priority = int(priority)
                else:
                    priority = 1
                func = ep.load()
                self.register(interface_package, func_name, func, priority)
            self._has_entry_point_lookup.add(key)
        return self._registered.get(key, None)

    def highest_priority(self, interface_package: str, func_name: str) -> Optional[Callable]:
        """Highest priority registered function with priority > 0."""
        registered = self.lookup(interface_package, func_name)
        if registered is None:
            return None
        highest = max(
            self._registered[(interface_package, func_name)],
            key=lambda x: self._priorities[x],
        )
        if self._priorities[highest] < 1:
            return None
        return highest

    def set_priority(self, func: Callable, priority: int) -> None:
        if func not in self._priorities:
            raise ValueError(f"Function {func} has not been registered")
        self._priorities[func] = priority

    def get_priority(self, func: Callable) -> int:
        if func not in self._priorities:
            raise ValueError(f"Function {func} has not been registered")
        return self._priorities[func]

    def disable(self, interface_package: str, func_name: str):
        """Set the priorites of all registered functions to -1."""
        registered = self.lookup(interface_package, func_name)
        for func in registered:
            self._priorities[func] = -1


function_factory = FunctionFactory()


def environment_dispatch(interface_package: str, func_name: str) -> Callable:
    factory_func = function_factory.highest_priority(interface_package, func_name)
    if factory_func is not None:
        return factory_func

    if sys.platform != "emscripten":
        if func_name.endswith("_async"):
            raise ValueError("async function are only implemented for emscripten")
        package = f"{interface_package}_wasi"
    else:
        if not func_name.endswith("_async"):
            raise ValueError("emscripten only implements the _async version of this function")
        package = f"{interface_package}_emscripten"
    mod = importlib.import_module(package)
    func = getattr(mod, func_name)

    return func



from typing import Optional, Union
from dataclasses import asdict
import copy

import numpy as np

from .image import Image, ImageType
from .pixel_types import PixelTypes
from .int_types import IntTypes
from .float_types import FloatTypes


def cast_image(
    input_image: Image,
    pixel_type: Optional[PixelTypes] = None,
    component_type: Optional[Union[IntTypes, FloatTypes]] = None,
) -> Image:
    """Cast an image to another pixel type and / or component type.

    :param input_image: Image to be cast.
    :type  input_image: Image

    :param pixel_type: Pixel type to cast to.
    :type  pixel_Type: Optional[PixelTypes]

    :param component_type: Component type to cast to.
    :type  component_type: Optional[Union[IntTypes, FloatTypes]]

    :return: Cast image. Event if pixel_type or component_type are not specified, a copy is made.
    :rtype:  Image

    """
    output_image_type = ImageType(**asdict(input_image.imageType))

    if pixel_type is not None:
        output_image_type.pixelType = pixel_type
        if pixel_type == PixelTypes.Scalar and output_image_type.components != 1:
            raise ValueError("PixelType Scalar requires components == 1")

    if component_type is not None and component_type != input_image.imageType.componentType:
        output_image_type.componentType = component_type

    output_image = Image(output_image_type)

    output_image.name = input_image.name
    output_image.origin = list(input_image.origin)
    output_image.spacing = list(input_image.spacing)
    output_image.direction = input_image.direction.copy()
    output_image.size = list(input_image.size)
    output_image.metadata = copy.deepcopy(input_image.metadata)

    if input_image.data is not None:
        if output_image_type.componentType == input_image.imageType.componentType:
            output_image.data = input_image.data.copy()
        else:
            component_type = output_image_type.componentType
            if component_type == IntTypes.UInt8:
                output_image.data = input_image.data.astype(np.uint8)
            elif component_type == IntTypes.Int8:
                output_image.data = input_image.data.astype(np.int8)
            elif component_type == IntTypes.UInt16:
                output_image.data = input_image.data.astype(np.uint16)
            elif component_type == IntTypes.Int16:
                output_image.data = input_image.data.astype(np.int16)
            elif component_type == IntTypes.UInt32:
                output_image.data = input_image.data.astype(np.uint32)
            elif component_type == IntTypes.Int32:
                output_image.data = input_image.data.astype(np.int32)
            elif component_type == IntTypes.UInt64:
                output_image.data = input_image.data.astype(np.uint64)
            elif component_type == IntTypes.Int64:
                output_image.data = input_image.data.astype(np.int64)
            elif component_type == FloatTypes.Float32:
                output_image.data = input_image.data.astype(np.float32)
            elif component_type == FloatTypes.Float64:
                output_image.data = input_image.data.astype(np.float64)
            else:
                raise ValueError("Unsupported component type")

    return output_image



from pathlib import Path

import itk

from itkwasm import Mesh
from dataclasses import asdict
import numpy as np


def test_mesh():
    data = Path(__file__).absolute().parent / "input" / "cow.vtk"
    itk_mesh = itk.meshread(data)

    itk_mesh_dict = itk.dict_from_mesh(itk_mesh)
    itkwasm_mesh = Mesh(**itk_mesh_dict)
    itkwasm_mesh_dict = asdict(itkwasm_mesh)
    itk_mesh_roundtrip = itk.mesh_from_dict(itkwasm_mesh_dict)
    itk_mesh_roundtrip_dict = itk.dict_from_mesh(itk_mesh_roundtrip)

    meshType = itk_mesh_dict["meshType"]
    meshType_roundtrip = itk_mesh_roundtrip_dict["meshType"]
    assert meshType["dimension"] == meshType_roundtrip["dimension"]

    assert meshType["pointComponentType"] == meshType_roundtrip["pointComponentType"]
    assert meshType["pointPixelComponentType"] == meshType_roundtrip["pointPixelComponentType"]
    assert meshType["pointPixelType"] == meshType_roundtrip["pointPixelType"]
    assert meshType["pointPixelComponents"] == meshType_roundtrip["pointPixelComponents"]

    assert meshType["cellComponentType"] == meshType_roundtrip["cellComponentType"]
    assert meshType["cellPixelComponentType"] == meshType_roundtrip["cellPixelComponentType"]
    assert meshType["cellPixelType"] == meshType_roundtrip["cellPixelType"]
    assert meshType["cellPixelComponents"] == meshType_roundtrip["cellPixelComponents"]

    assert itk_mesh_dict["name"] == itk_mesh_roundtrip_dict["name"]

    assert itk_mesh_dict["numberOfPoints"] == itk_mesh_roundtrip_dict["numberOfPoints"]
    assert np.array_equal(itk_mesh_dict["points"], itk_mesh_roundtrip_dict["points"])

    assert itk_mesh_dict["numberOfPointPixels"] == itk_mesh_roundtrip_dict["numberOfPointPixels"]
    assert np.array_equal(itk_mesh_dict["pointData"], itk_mesh_roundtrip_dict["pointData"])

    assert itk_mesh_dict["numberOfCells"] == itk_mesh_roundtrip_dict["numberOfCells"]
    assert np.array_equal(itk_mesh_dict["cells"], itk_mesh_roundtrip_dict["cells"])
    assert itk_mesh_dict["cellBufferSize"] == itk_mesh_roundtrip_dict["cellBufferSize"]

    assert itk_mesh_dict["numberOfCellPixels"] == itk_mesh_roundtrip_dict["numberOfCellPixels"]
    assert np.array_equal(itk_mesh_dict["cellData"], itk_mesh_roundtrip_dict["cellData"])



from pathlib import Path

import itk

from itkwasm import Image
from dataclasses import asdict
import numpy as np


def test_image():
    data = Path(__file__).absolute().parent / "input" / "cthead1.png"
    itk_image = itk.imread(data, itk.UC)
    itk_image_dict = itk.dict_from_image(itk_image)
    itkwasm_image = Image(**itk_image_dict)
    itkwasm_image_dict = asdict(itkwasm_image)
    itk_image_roundtrip = itk.image_from_dict(itkwasm_image_dict)
    difference = np.sum(itk.comparison_image_filter(itk_image, itk_image_roundtrip))
    assert difference == 0.0


def test_image_defaults():
    image = Image()

    assert image.imageType.dimension == 2
    assert image.imageType.componentType == "uint8"
    assert image.imageType.pixelType == "Scalar"
    assert image.imageType.components == 1

    assert image.name == "Image"
    assert image.origin[0] == 0.0
    assert image.origin[1] == 0.0
    assert image.spacing[0] == 1.0
    assert image.spacing[1] == 1.0
    assert np.array_equal(image.direction, np.eye(2).astype(np.float64))

    assert image.size[0] == 1
    assert image.size[1] == 1

    assert isinstance(image.metadata, dict)
    assert image.data == None



import sys

if sys.version_info < (3, 10):
    from importlib_metadata import distribution
else:
    from importlib.metadata import distribution

from numpy.typing import ArrayLike

try:
    import cupy

    _CUPY_AVAILABLE = True
except:
    _CUPY_AVAILABLE = False

try:
    distribution("dask")
    _DASK_AVAILABLE = True
except:
    _DASK_AVAILABLE = False


def is_cupy_array(arr: ArrayLike) -> bool:
    """Check if the input is a CuPy array.

    :param arr: Input array
    :type  arr: ArrayLike

    :return: True if the input is a CuPy array
    :rtype:  bool
    """
    if not _CUPY_AVAILABLE:
        return False
    import cupy as cp

    return isinstance(arr, cp.ndarray)


def array_like_to_cupy_array(arr: ArrayLike) -> "cp.ndarray":
    """Convert a numpy array-like to a cupy ndarray.

    :param arr: numpy ndarray like
    :type  arr: ArrayLike

    :return: cupy array
    :rtype:  cp.ndarray
    """
    import cupy as cp

    if isinstance(arr, cp.ndarray):
        return arr
    if _DASK_AVAILABLE:
        import dask.array as da

        if isinstance(arr, da.Array):
            arr = arr.compute()
        if isinstance(arr, cp.ndarray):
            return arr
    return cp.array(arr)



import numpy as np

from itkwasm import image_from_array, ImageType, PixelTypes, IntTypes, FloatTypes


def test_image_from_array():
    arr = np.random.rand(10, 10)
    image = image_from_array(arr)
    assert np.array_equal(arr, image.data)
    assert image.imageType.dimension == 2
    assert image.imageType.componentType == FloatTypes.Float64
    assert image.imageType.pixelType == PixelTypes.Scalar
    assert image.imageType.components == 1


def test_image_from_array_anisotropic():
    arr = np.random.rand(10, 5)
    image = image_from_array(arr)
    assert np.array_equal(arr, image.data)
    assert image.imageType.dimension == 2
    assert image.imageType.componentType == FloatTypes.Float64
    assert image.imageType.pixelType == PixelTypes.Scalar
    assert image.imageType.components == 1
    assert image.size[0] == 5
    assert image.size[1] == 10


def test_image_from_array_vector():
    arr = np.random.rand(10, 10, 3)
    image = image_from_array(arr, is_vector=True)
    assert np.array_equal(arr, image.data)
    assert image.imageType.dimension == 2
    assert image.imageType.componentType == FloatTypes.Float64
    assert image.imageType.pixelType == PixelTypes.VariableLengthVector
    assert image.imageType.components == 3


def test_image_from_array_explicit():
    arr = np.random.rand(10, 10, 3)
    image_type = {
        "dimension": 2,
        "componentType": FloatTypes.Float32,
        "pixelType": PixelTypes.VariableLengthVector,
        "components": 3,
    }
    image_type = ImageType(**image_type)
    image = image_from_array(arr, image_type=image_type)
    assert np.array_equal(arr, image.data)
    assert image.imageType.dimension == 2
    assert image.imageType.componentType == FloatTypes.Float32
    assert image.imageType.pixelType == PixelTypes.VariableLengthVector
    assert image.imageType.components == 3



from typing import Optional

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike

from .image import Image, ImageType
from .to_numpy_array import _dtype_to_component_type

from .pixel_types import PixelTypes


def image_from_array(arr, is_vector: bool = False, image_type: Optional[ImageType] = None) -> Image:
    """Convert a numpy array-like to an itkwasm Image.

    :param arr: Input n-d array
    :type  arr: NumPy ndarray or array-like, e.g. dask.Array or cupy.ndarray

    :param is_vector: Whether the array is a vector image. If True, then a 3D array will be treated as a 2D vector image, otherwise it will be treated as a 3D image.
    :type  is_vector: bool

    :param image_type: Explicitly specify the Image type
    :type  image_type: ImageType

    :return: Output Image
    :rtype : Image where pixel data is a view of the input array and default metadata
    """

    dimension = arr.ndim
    if is_vector:
        dimension -= 1

    if image_type is None:
        image_type = ImageType(
            dimension=dimension,
            componentType=_dtype_to_component_type(arr.dtype),
            pixelType=(PixelTypes.Scalar if not is_vector else PixelTypes.VariableLengthVector),
            components=arr.shape[-1] if is_vector else 1,
        )

    image = Image(imageType=image_type)
    image.size = arr.shape[:dimension][::-1]
    image.data = arr

    return image



from dataclasses import dataclass, field

from typing import Sequence, Union, Dict, Optional

try:
    from numpy.typing import ArrayLike
except ImportError:
    from numpy import ndarray as ArrayLike
import numpy as np

from .int_types import IntTypes
from .float_types import FloatTypes
from .pixel_types import PixelTypes


@dataclass
class ImageType:
    dimension: int = 2
    componentType: Union[IntTypes, FloatTypes] = IntTypes.UInt8
    pixelType: PixelTypes = PixelTypes.Scalar
    components: int = 1


def _default_direction() -> ArrayLike:
    return np.empty((0,), np.float64)


@dataclass
class ImageRegion:
    index: Sequence[int] = field(default_factory=list)
    size: Sequence[int] = field(default_factory=list)


@dataclass
class Image:
    imageType: Union[ImageType, Dict] = field(default_factory=ImageType)
    name: str = "Image"
    origin: Sequence[float] = field(default_factory=list)
    spacing: Sequence[float] = field(default_factory=list)
    direction: ArrayLike = field(default_factory=_default_direction)
    size: Sequence[int] = field(default_factory=list)
    metadata: Dict = field(default_factory=dict)
    data: Optional[ArrayLike] = None
    bufferedRegion: Optional[ImageRegion] = None

    def __post_init__(self):
        if isinstance(self.imageType, dict):
            self.imageType = ImageType(**self.imageType)

        dimension = self.imageType.dimension
        if len(self.origin) == 0:
            self.origin += [
                0.0,
            ] * dimension

        if len(self.spacing) == 0:
            self.spacing += [
                1.0,
            ] * dimension

        if len(self.direction) == 0:
            self.direction = np.eye(dimension).astype(np.float64)

        if len(self.size) == 0:
            self.size += [
                1,
            ] * dimension

        if self.bufferedRegion is None:
            self.bufferedRegion = ImageRegion(
                index=[
                    0,
                ]
                * dimension,
                size=self.size,
            )
